struct entity_t;
// Different types of call-back functions for each entity:
typedef void (*func_t(struct entity_t *self));
typedef void (*touch_t(struct entity_t *self, struct entity_t *other));

// Entity struct:
typedef struct entity_t {
	unsigned char	index;
	unsigned short	class;
	
	// To retrieve actual screen position, divide by 256 ( >> 8)
	signed long		x, y;
	signed long		oldx, oldy;
	signed long		dx, dy;		// Scaled up by 256 ( >> 8)
	
	unsigned char	colflags;
	
	unsigned long	flags;
	unsigned char	frame;
	unsigned char	width, height;
	
	struct entity_t	*owner;

	signed long		next_think;		// Next time that this entity will execute think()
	func_t			*think;

	// Im not so sure I think I need the render() function
	// to be dependent upon the entity's needs, but I'll leave it in.
	func_t			*render;
	func_t			*death;
	touch_t			*touch;

	// Extra variables:

	signed long		attack_time;	// Next time that this entity is allowed to attack
	signed short	health;
	signed short	max_health;
	
	signed short	attack_damage;
} entity_t;

// Flags for the flags property of entities:
#define FLAG_FACERIGHT		0x00000001
#define FLAG_ONGROUND		0x00000002
#define FLAG_LIEDOWN		0x00000004
#define FLAG_LOOKUP			0x00000008
#define FLAG_NOGRAVITY		0x00000010
#define FLAG_NOFRICTION		0x00000020
#define FLAG_DEAD			0x00000040
#define FLAG_BULLETPIERCE	0x00001000

// Flags returned thru check_map(entity_t *) function:
#define	BLOCKED_GROUND		0x01
#define BLOCKED_LHS			0x02
#define BLOCKED_RHS			0x04
#define BLOCKED_TOP			0x08

#define CLASS_NONE			0
#define CLASS_PLAYER		1
#define CLASS_GRUNT			2
#define CLASS_BULLET		3

#define	MAX_ENTITIES		64

entity_t		**entities = NULL;		// entities on one screen at a time:
entity_t		*player = NULL;
unsigned int	time = 0;
long	screen_x = 0, screen_y = 0;

unsigned char	*map = NULL;			// Tile-Map pointer.
unsigned char	horiz_scroll = TRUE;
signed char		grayadjust = 0;

void *safe_malloc(size_t size) {
	void *ptr = NULL;
	ptr = malloc(size);
	return ptr;
};


entity_t *spawn() {
	unsigned char i;
	for (i=0; i<MAX_ENTITIES; i++)
		if (entities[i] == NULL) break;
	if (i == MAX_ENTITIES) return NULL;
	
	entity_t	*e = safe_malloc(sizeof(entity_t));
	if (e == NULL) {
		printf_xy(0, 0, "spawn(): Not enough memory!");	
		return NULL;
	}
	entities[i] = e;
	e->index = i;

	// Initialize all the data:
	e->oldx = e->oldy = e->x = e->y = e->dx = e->dy = 0;
	e->class = CLASS_NONE;
	e->frame = 0;
	e->width = 8; e->height = 8;
	e->flags = FLAG_FACERIGHT;
	e->owner = NULL;
	e->render = NULL;
	e->think = NULL;
	e->touch = NULL;
	e->death = NULL;
	e->attack_time = 0;
	e->next_think = 0;
	e->attack_damage = 0;
	e->health = 256;
	e->max_health = 256;
	
	return e;
};

void kill(entity_t *e) {
	if (e == NULL) return;
	if ((e->index > MAX_ENTITIES) || (e < 0)) return;
	entities[e->index] = NULL;
	free(e);
	e = NULL;
};

void damage(entity_t *target, entity_t *inflictor, entity_t *attacker, signed short dmg) {
	if (target == NULL || inflictor == NULL || attacker == NULL) return;
	
	target->health -= dmg;
	if (target->health > target->max_health) target->health = target->max_health;
	if (target->health < 0) {
		// Target has died!
		if (target->death) target->death(target);
	};
};


// Set this to think function for entity to remove it: (e.g. Quake-style)
void sub_remove(struct entity_t *self) {
	if (self == NULL) return;
	kill(self);
};

// Test if the two boxes collide:
bool inside_test(int xoo, int yoo, int xoe, int yoe, int ow, int oh, int ew, int eh) {
	return (((xoe >= xoo) && (xoe <= xoo + ow) && (yoe >= yoo) && (yoe <= yoo + oh)) ||
	    	((xoo >= xoe) && (xoo <= xoe + ew) && (yoo >= yoe) && (yoo <= yoe + eh)));/* ||
			((xoo + ow >= xoe) && (xoo + ow <= xoe + ew) && 
				(yoo >= ));*/
};

// This function assumes the pointers are valid and are not NULL:
// Perform this check before calling the function.
bool collide(entity_t *e, entity_t *o) {
	bool test = FALSE;
	short xoe, yoe, xoo, yoo;
	int i, j;
	
	if (e == NULL || o == NULL) return FALSE;
	
	// Check for a collision between entity E and O:
	// If they're close enough, do a sprite-compare:
	xoe = e->x >> 8;
	xoo = o->x >> 8;
	yoe = e->y >> 8;
	yoo = o->y >> 8;
	
	#define ow (o->width)
	#define oh (o->height)
	#define ew (e->width)
	#define eh (e->height)
	
	if (((xoe >= xoo) && (xoe <= xoo + ow) && (yoe >= yoo) && (yoe <= yoo + oh)) ||
	    ((xoo >= xoe) && (xoo <= xoe + ew) && (yoo >= yoe) && (yoo <= yoe + eh)))
	{
		return TRUE;
	}

	#undef ow
	#undef oh
	#undef ew
	#undef eh
	
	return test;
};

/*
unsigned char check_map(entity_t *curr) {
	bool bot_check, rhs_check, lhs_check, top_check;
	unsigned char total = 0;
	int tx, ty, sx, sy, w, h, i;
	
	unsigned long access;
	
	// {sx, sy} are the pixel coordinates of the current thing.
	sx = (curr->x >> 8);
	sy = (curr->y >> 8);
	
	// {tx, ty} are the tile-coordinates of the current thing on the map.
	tx = (curr->x >> 11) % TILE_W;
	ty = (curr->y >> 11) % TILE_H;

	w = (curr->width >> 3); h = (curr->height >> 3);
	
	bot_check = FALSE;
	lhs_check = FALSE;
	rhs_check = FALSE;
	top_check = FALSE;

	// Bottom check:
	access = ( TILE_W * (ty + h - 1) ) + (tx);
	if ( ((map[access] & 0x80) == 0x80) && 
	     inside_test(sx, sy, tx << 3, (ty + h) << 3, curr->width, curr->height, 8, 8) )
	{
		// Entity collides with the tile...
		bot_check = TRUE;
	};
	
	total = 0;
	total |= (bot_check ? BLOCKED_GROUND : 0x00);
	total |= (lhs_check ? BLOCKED_LHS : 0x00);
	total |= (rhs_check ? BLOCKED_RHS : 0x00);
	total |= (top_check ? BLOCKED_TOP : 0x00);
		
	return total;
};
*/

unsigned char check_map(entity_t *curr) {
	bool ground_check, rhs_check, lhs_check, top_check;
	unsigned char total = 0;
	int tx, ty, w, h, i;
	
	unsigned long access;
	
	// {tx, ty} are the tile-coordinates of the current thing on the map.
	tx = (curr->x >> 11) % TILE_W;
	ty = (curr->y >> 11) % TILE_H;

	w = (curr->width >> 3); h = (curr->height >> 3);
	
	ground_check = FALSE;
	if (curr->flags & FLAG_FACERIGHT) {
		for (i = 0; i < w; i++) {
			access = (TILE_W * (ty + h)) + tx + i;
			ground_check |= ( (map[ access ] & 0x80) == 0x80 );
		}
	} else {
		for (i = 0; i < w; i++) {
			access = (TILE_W * (ty + h)) + tx - i;
			ground_check |= ( (map[ access ] & 0x80) == 0x80 );
		}
	}
	
	lhs_check = FALSE;
	for (i = 0; i < h; i++) {
		access = (TILE_W * (ty + i)) + tx - ((curr->flags & FLAG_FACERIGHT) ? 1 : w);
		lhs_check |= ( (map[ access ] & 0x80) == 0x80 );
	}

	rhs_check = FALSE;
	for (i = 0; i < h; i++) {
		access = (TILE_W * (ty + i)) + tx + ((curr->flags & FLAG_FACERIGHT) ? w : 0);
		rhs_check |= ( (map[ access ] & 0x80) == 0x80 );
	}
	
	top_check = FALSE;
	
	total = 0;
	total |= (ground_check ? BLOCKED_GROUND : 0x00);
	total |= (lhs_check ? BLOCKED_LHS : 0x00);
	total |= (rhs_check ? BLOCKED_RHS : 0x00);
	total |= (top_check ? BLOCKED_TOP : 0x00);
	
	return total;
};

